#![allow(dead_code)]

mod re;
mod text;
mod args;
mod ext; use ext::*;

use color_eyre::{
    eyre::{
        self,
        eyre,
        WrapErr as _,
    },
    SectionExt as _, Help as _,
};

fn initialise() -> eyre::Result<()>
{
    color_eyre::install()?;
    Ok(())
}

/// Provides the group output writer with references to the byte-strings to be used to delimit single groups' output & to delimit each input string (or line)'s output.
///
/// The defaults are `\n` (LF-literal) for [line_delimiter], and `\t` for [field_delimiter], see the fields for more information.
#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord/*, Copy*/)]
pub struct GroupConfig<'a>
{
    /// Delimits the output of each individual input string.
    ///
    /// By default, this is a newline `\n` character.
    pub line_delimiter: &'a [u8],
    /// Delimits the output of each requested capture group (per [line_delimiter].)
    /// If there is only 1 (or less) requested groups (that may be *empty/unmatched*, are **not** *invalid*,)
    /// then there is no field delimit for that individual output string line.
    ///
    /// By default, this is a tab literal `\t` character.
    pub field_delimiter: &'a [u8],
}

impl<'a> Default for GroupConfig<'a>
{
    #[inline]
    fn default() -> Self
    {
	Self::new()
    }
}

impl<'a> GroupConfig<'a>
{
    pub const fn new() -> Self {
	Self {
	    line_delimiter: b"\n",
	    field_delimiter: b"\t",
	}
    }

    pub const fn has_line_delimiter(&self) -> bool {
	! self.line_delimiter.is_empty()
    }
    pub const fn has_field_delimiter(&self) -> bool {
	! self.field_delimiter.is_empty()
    }
    
    #[inline] 
    pub const fn with_field_delimiter<'b>(self, field_delimiter: &'b [u8]) -> GroupConfig<'b>
    where 'a: 'b
    {
	GroupConfig {
	    field_delimiter,
	    ..self
	}
    }
    #[inline] 
    pub const fn with_field_delimiter_str<'b>(self, field: &'b str) -> GroupConfig<'b>
    where 'a: 'b
    {
	GroupConfig {
	    field_delimiter: field.as_bytes(),
	    ..self
	}
    }
    #[inline] 
    pub const fn with_line_delimiter<'b>(self, line_delimiter: &'b [u8]) -> GroupConfig<'b>
    where 'a: 'b
    {
	GroupConfig {
	    line_delimiter,
	    ..self
	}
    }
    #[inline] 
    pub const fn with_line_delimiter_str<'b>(self, line: &'b str) -> GroupConfig<'b>
    where 'a: 'b
    {
	GroupConfig {
	    line_delimiter: line.as_bytes(),
	    ..self
	}
    }
}

#[inline] 
fn print_groups<'c, 'a, S: ?Sized, G, T: 'a, I>(to: &mut S, g: G, groups: I, how: impl std::borrow::Borrow<GroupConfig<'c>>) -> std::io::Result<()>
where S: std::io::Write + 'c, // NOTE: This lifetime bound is not yet used, as it is just `Write`, but if we change this to a context wrapper, then we can copy the `how`'s `'c` references into the context object without direct write/format/cloning.
      G: IntoIterator<Item = &'a Option<T>> + Clone + Copy, // NOTE: Copy bound to ensure we're not accidentally doing deep clones of `g`.
//G: std::ops::Index<usize>, G::Output: std::borrow::Borrow<Option<T>>,
      T: std::borrow::Borrow<str>,
      I: IntoIterator<Item: std::borrow::Borrow<usize>/*, IntoIter: ExactSizeIterator*/>,
{
    use std::{
	borrow::Borrow,
	io::Write,
    };
    let how = how.borrow(); //std::borrow::ToOwned::clone_into(&self, target);
    let mut first = true;
    for group in groups.into_iter() {
	let group = group.borrow();
	// // Moved to into match group (skipping invalid groups.)
	// if !first {
	//     write!(to, "\t")?;
	// }
	let print_delim = move |to: &mut S| to.write_all(first.then_some(&[][..]).unwrap_or(&how.field_delimiter[..]).as_ref()); // If it's not the first iteration, print `\t`.
	match g.into_iter().nth(*group) {
	    Some(None) => print_delim(to),
	    Some(Some(g)) => {
		print_delim(to)?;
		write!(to, "{}", g.borrow())
	    },

	    //TODO: What should be the behaviour of a non-existent group index here? (NOTE: This now corresponds to the previous `g.len() > group` check in caller.) // (NOTE: The original behaviour is to just ignore groups that are out of range entirely (i.e. no printing, no delimit char, no error,) maybe treat non-existent groups as non-matched groups and *just* print the delim char?)
	    // (NOTE: Moved out of branch, see above ^) // None if !first => write!(to, "\t"),
	    // XXX: Should this do what it does now...? Or should it `break` to prevent the checking for more groups...? Print a warning maybe...?
	    None => {
		eprintln!("Warning: Invalid group index {}!", group);
		continue; // Do not set `first = false` if it was an invalid index.
		//Ok(())
	    }, 
	}?;
	first = false;
    }
    // If `first == true`, no groups were printed, so we do not print the new-line.
    if !first && how.has_line_delimiter() {
	to.write_all(how.line_delimiter.as_ref())
    } else {
	Ok(())
    }
}

fn main() -> eyre::Result<()>
{
    initialise().wrap_err("Fatal: Failed to install panic handle")?;
    
    //let cli = args::parse_cli();//.wrap_err("Error parsing command-line arguments")?;
    
    //eprintln!("{:#?}", cli);
    // return Ok(());

    let args: re::FrozenVec<re::FrozenString> = std::env::args().map(String::into_boxed_str).collect();

    if args.len() < 4 {
	use owo_colors::OwoColorize;
	println!("rematch v{}: Regular-expression group matcher", env!("CARGO_PKG_VERSION"));
	println!("");
	println!("Usage: {} <str> <regex> <group>...", args[0]);
	println!("Pass `-' as `<str>' to read lines from stdin");
	println!("");
	println!("Enabled Features:");
	if cfg!(feature="perl") {
	    println!("{}\t\t\tEnable PCRE2 (extended) regular-expressions.\n\t\t\tNote that PCRE2 regex engine matches on *bytes*, not *characters*; meaning if a match cuts a vlid UTF8 codepoint into an invalid one, the output will replace the invalid characters with U+FFFD REPLACEMENT CHARACTER.", disjoint!["+", "perl"].red());
	} else {
	    println!("{}\t\t\tPCRE2 (extended) features are disabled; a faster but less featureful regular expression engine (that matches on UTF8 strings instead of raw bytes) is used instead.", disjoint!["-", "perl"].bright_blue());
	}
	if cfg!(feature="unstable") {
	    println!("+unstable\t\tUnstable optimisations evailable & enabled for build.");
	} else {
	    println!("-unstable\t\tUnstable optimisations disabled / not available for build.");
	}
	std::process::exit(1)
    } else {
	let re = re::Regex::compile(&args[2])?;
	let text = &args[1];

	let print_cfg = GroupConfig::new();
	let groups = {
	    let groups = &args[3..];

	    if groups.len() < 1 {
		eprintln!("Warning: No capture groups requested.");
		// NOTE: Unexpected branch...
		return Ok(());
	    }

	    // Parse each group index into `groups`.
	    groups.iter().enumerate()
		.map(|(i, x)| x.parse()
		     .with_section(|| format!("{:?}", groups).header("Groups specified were"))
		     .with_section(|| x.clone().header("Specified capture group index was"))
		     .with_section(move || i.header("Argument index in provided groups")))
		.collect::<Result<Box<[usize]>, _>>()
		.wrap_err("Invalid group index specified")?
	};

	use std::io::Write;
	let mut stdout = std::io::stdout();

	// Take the kind of `stdout` used (locked & buffered, or not locked & buffered) ..
	let stdout = if &text[..] == "-" {
	    // Lock the output for the duration of the read lines.
	    // Buffer the output in program memory to make processing a bit faster (i.e. the segmented 'write' operations in `print_groups()`, which may be called many times here) & not have to wait on write lines for no reason (since we're already waiting on read lines.)
	    let mut stdout = std::io::BufWriter::new(stdout.lock());
	    
	    text::stdin_lines(|text| -> eyre::Result<bool> {
		match re.exec(&text)? {
		    Some(g) /*if g.len() > group*/ => // NOTE: This check branch has now been moved into `print_groups()`
			print_groups(&mut stdout, &g, &groups, &print_cfg)?, //println!("{}", &g[group]),
		    _ => (),
		}
		Ok(true)
	    })?;

	    // Return the buffer to the main block to be flushed to output (see above & below.)
	    Some(stdout)
	} else {
	    match re.exec(&text)? {
		Some(g) /*if g.len() > group*/ => print_groups(&mut stdout, &g[..], &groups, print_cfg)?,//println!("{}", &g.nth(group).unwrap().map(|x| x.as_ref()).unwrap_or("")),
		_ => (),
	    }

	    // As there is no internal buffer used, there is no reason to pass it to be flushed by the program when finished like the `-` case above.
	    None
	}.ok_or_else(move || stdout); // NOTE ^: Instead, we have it flush the non-memory-buffered output handle before exiting the program.
	
	// and .. Ensure the stream (and buffer, if used) is flushed (then dropped.)
	unwrap_either!(mut stdout => stdout.flush()).unwrap();
    }
    Ok(())
}
